import torch
import random
import numpy as np
from torch import nn
from torch import Tensor
from torchvision.datasets import MNIST, CIFAR10
from torch.utils.data import DataLoader
from torchvision import transforms as T
import torch.optim as optim
from typing import Type, Any, Callable, Union, List, Optional
import torch.nn.functional as F
from torch.autograd import Variable
from argparse import ArgumentParser
from wideresnet import *
import os

def _pgd_whitebox(model,
                  X,
                  y,
                  device,
                  epsilon,
                  num_steps,
                  step_size,
                  random):
    out = model(X)
    pred = out.data.max(1, keepdim=True)[1]
    err = (out.data.max(1)[1] != y.data).float().sum()
    X_pgd = Variable(X.data, requires_grad=True)
    if random:
        random_noise = torch.FloatTensor(*X_pgd.shape).uniform_(-epsilon, epsilon).to(device)
        X_pgd = Variable(X_pgd.data + random_noise, requires_grad=True)

    for _ in range(num_steps):
        opt = optim.SGD([X_pgd], lr=1e-3)
        opt.zero_grad()

        with torch.enable_grad():
            loss = nn.CrossEntropyLoss()(model(X_pgd), y)
        loss.backward()
        eta = step_size * X_pgd.grad.data.sign()
        X_pgd = Variable(X_pgd.data + eta, requires_grad=True)
        eta = torch.clamp(X_pgd.data - X.data, -epsilon, epsilon)
        X_pgd = Variable(X.data + eta, requires_grad=True)
        X_pgd = Variable(torch.clamp(X_pgd, 0, 1.0), requires_grad=True)
    err_pgd = (model(X_pgd).data.max(1)[1] != y.data).float().sum()
    output = model(X_pgd)
    adv_pred = output.data.max(1, keepdim=True)[1]

    print('err pgd (white-box): ', err_pgd)
    return err, err_pgd, pred, adv_pred


def _pgd_blackbox(model_target,
                  model_source,
                  X,
                  y,
                  device,
                  epsilon,
                  num_steps,
                  step_size,
                  random):
    out = model_target(X)
    pred = out.data.max(1, keepdim=True)[1]
    err = (out.data.max(1)[1] != y.data).float().sum()
    X_pgd = Variable(X.data, requires_grad=True)
    if random:
        random_noise = torch.FloatTensor(*X_pgd.shape).uniform_(-epsilon, epsilon).to(device)
        X_pgd = Variable(X_pgd.data + random_noise, requires_grad=True)

    for _ in range(num_steps):
        opt = optim.SGD([X_pgd], lr=1e-3)
        opt.zero_grad()

        with torch.enable_grad():
            loss = nn.CrossEntropyLoss()(model_source(X_pgd), y)
        loss.backward()
        eta = step_size * X_pgd.grad.data.sign()
        X_pgd = Variable(X_pgd.data + eta, requires_grad=True)
        eta = torch.clamp(X_pgd.data - X.data, -epsilon, epsilon)
        X_pgd = Variable(X.data + eta, requires_grad=True)
        X_pgd = Variable(torch.clamp(X_pgd, 0, 1.0), requires_grad=True)

    err_pgd = (model_target(X_pgd).data.max(1)[1] != y.data).float().sum()
    output = model_target(X_pgd)
    adv_pred = output.data.max(1, keepdim=True)[1]

    print('err pgd black-box: ', err_pgd)
    return err, err_pgd, pred, adv_pred


def eval_adv_test_whitebox(model, device, test_loader, args):
    """
    evaluate model by white-box attack
    """
    model.eval()
    robust_err_total = 0
    natural_err_total = 0
    correct = 0
    correct_adv = 0
    for data, target in test_loader:
        data, target = data.to(device), target.to(device)
        # pgd attack
        X, y = Variable(data, requires_grad=True), Variable(target)
        err_natural, err_robust, pred, adv_pred = _pgd_whitebox(model, X, y, device, args.epsilon, args.attack_num_steps, args.attack_step_size, args.random)
        correct += pred.eq(target.data.view_as(pred)).sum()
        correct_adv += adv_pred.eq(target.data.view_as(adv_pred)).sum()
        robust_err_total += err_robust
        natural_err_total += err_natural
    print('natural_err_total: ', natural_err_total)
    print('robust_err_total: ', robust_err_total)
    print('clean accuracy: {}/{} ({:.4f}%)'.format(correct, len(test_loader.dataset), 100 * correct/len(test_loader.dataset)))
    print('adv accuracy: {}/{} ({:.4f}%)'.format(correct_adv, len(test_loader.dataset), 100 *correct_adv/len(test_loader.dataset)))


def eval_adv_test_blackbox(model_target, model_source, device, test_loader, args):
    """
    evaluate model by black-box attack
    """
    model_target.eval()
    model_source.eval()
    robust_err_total = 0
    natural_err_total = 0
    correct = 0
    correct_adv = 0
    for data, target in test_loader:
        data, target = data.to(device), target.to(device)
        # pgd attack
        X, y = Variable(data, requires_grad=True), Variable(target)
        err_natural, err_robust, pred, adv_pred = _pgd_blackbox(model_target, model_source, X, y, device, args.epsilon, args.attack_num_steps, args.attack_step_size, args.random)
        correct += pred.eq(target.data.view_as(pred)).sum()
        correct_adv += adv_pred.eq(target.data.view_as(adv_pred)).sum()
        robust_err_total += err_robust
        natural_err_total += err_natural
    print('natural_err_total: ', natural_err_total)
    print('robust_err_total: ', robust_err_total)
    print('clean accuracy: {}/{} ({:.4f}%)'.format(correct, len(test_loader.dataset), 100 * correct/len(test_loader.dataset)))
    print('adv accuracy: {}/{} ({:.4f}%)'.format(correct_adv, len(test_loader.dataset), 100 *correct_adv/len(test_loader.dataset)))


def fix_seeds(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

def get_dataloaders(train_batch_size, test_batch_size):    
    train_loader = torch.utils.data.DataLoader(
        CIFAR10('./data/CIFAR10/', train=True, download=True,
                    transform=T.Compose([
                        T.RandomCrop(32, padding=4),
                        T.RandomHorizontalFlip(),
                        T.ToTensor(),
                    ])),
        batch_size=train_batch_size, shuffle=True, num_workers=4, pin_memory=True)
    test_loader = torch.utils.data.DataLoader(
        CIFAR10('./data/CIFAR10/', train=False, transform=T.Compose([
            T.ToTensor(),
        ])),
        batch_size=test_batch_size, shuffle=False, num_workers=4, pin_memory=True)

    return train_loader, test_loader


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def adjust_learning_rate(optimizer, epoch, args):
    """decrease the learning rate"""
    lr = args.lr
    
    if epoch >= 75:
        lr = args.lr * 0.1
    if epoch >= 90:
        lr = args.lr * 0.01
    if epoch >= 100:
        lr = args.lr * 0.001
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr

def adv_loss(model,
                x_natural,
                y,
                optimizer,
                args,
                step_size=0.003,
                epsilon=0.031,
                perturb_steps=10,
                beta=1.0,
                distance='l_inf',
                TRADES=True):
    # define KL-loss
    criterion_kl = nn.KLDivLoss(reduction='sum')
    model.eval()
    batch_size = len(x_natural)
    # generate adversarial example
    x_adv = x_natural.detach() + 0.001 * torch.randn(x_natural.shape).cuda().detach()
    
    if distance == 'l_inf':
        for _ in range(perturb_steps):
            x_adv.requires_grad_()
            with torch.enable_grad():
                loss_kl = criterion_kl(F.log_softmax(model(x_adv), dim=1),
                                       F.softmax(model(x_natural), dim=1))
            grad = torch.autograd.grad(loss_kl, [x_adv])[0]
            x_adv = x_adv.detach() + step_size * torch.sign(grad.detach())
            x_adv = torch.min(torch.max(x_adv, x_natural - epsilon), x_natural + epsilon)
            x_adv = torch.clamp(x_adv, 0.0, 1.0)
    elif distance == 'l_2':
        delta = 0.001 * torch.randn(x_natural.shape).cuda().detach()
        delta = Variable(delta.data, requires_grad=True)

        # Setup optimizers
        optimizer_delta = optim.SGD([delta], lr=epsilon / perturb_steps * 2)

        for _ in range(perturb_steps):
            adv = x_natural + delta

            # optimize
            optimizer_delta.zero_grad()
            with torch.enable_grad():
                loss = (-1) * criterion_kl(F.log_softmax(model(adv), dim=1),
                                           F.softmax(model(x_natural), dim=1))
            loss.backward()
            # renorming gradient
            grad_norms = delta.grad.view(batch_size, -1).norm(p=2, dim=1)
            delta.grad.div_(grad_norms.view(-1, 1, 1, 1))
            # avoid nan or inf if gradient is 0
            if (grad_norms == 0).any():
                delta.grad[grad_norms == 0] = torch.randn_like(delta.grad[grad_norms == 0])
            optimizer_delta.step()

            # projection
            delta.data.add_(x_natural)
            delta.data.clamp_(0, 1).sub_(x_natural)
            delta.data.renorm_(p=2, dim=0, maxnorm=epsilon)
        x_adv = Variable(x_natural + delta, requires_grad=False)
    elif distance == 'vat':
        vat_loss = VATLoss(xi=args.xi, epsilon=args.eps, ip=args.ip)
        r_adv = vat_loss(model, x_natural)
        x_adv = Variable(x_natural + r_adv, requires_grad=False)
    else:
        x_adv = torch.clamp(x_adv, 0.0, 1.0)
    model.train()

    x_adv = Variable(torch.clamp(x_adv, 0.0, 1.0), requires_grad=False)
    # zero gradient
    optimizer.zero_grad()
    # calculate loss

    if TRADES:
        logits = model(x_natural)
        loss = loss_natural + beta * loss_robust
    else:
        logits = model(x_adv)
        loss = F.cross_entropy(logits, y)
    return loss

def adversarial_train(train_loader, model, device, criterion, optimizer, epoch, args):
    losses = AverageMeter()
    model.to(device)
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        output = model(data)
        loss = adv_loss(model=model,
                           x_natural=data,
                           y=target,
                           optimizer=optimizer,
                           args=args,
                           step_size=args.step_size,
                           epsilon=args.epsilon,
                           perturb_steps=args.num_steps,
                           distance=args.distance,
                           beta=args.beta,
                           TRADES=args.TRADES)

        losses.update(loss.item(), data.size(0))

        loss.backward()
        optimizer.step()

        if batch_idx % args.log_interval == 0:
            print('Adv Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), losses.avg))

    return losses.avg
    #   torch.save(network.state_dict(), '/results/model.pth')
    #   torch.save(optimizer.state_dict(), '/results/optimizer.pth')

def test(model, device, criterion, test_loader):
    model.to(device)
    model.eval()
    losses = AverageMeter()
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            loss = criterion(output, target)
            # test_loss += criterion(output, target).item()
            losses.update(loss.item(), data.size(0))

            pred = output.data.max(1, keepdim=True)[1]
            correct += pred.eq(target.data.view_as(pred)).sum()
        print('\nTest set: Avg. loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
            losses.avg, correct, len(test_loader.dataset),
            100. * correct / len(test_loader.dataset)))
    return losses.avg, correct / len(test_loader.dataset)

def train(train_loader, model, device, criterion, optimizer, epoch, args):
    losses = AverageMeter()
    model.to(device)
    model.train()

    for batch_idx, (data, target) in enumerate(train_loader):
        optimizer.zero_grad()
        data, target = data.to(device), target.to(device)
        output = model(data)
        loss = criterion(output, target)

        losses.update(loss.item(), data.size(0))

        loss.backward()
        optimizer.step()

        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), losses.avg))
    return losses.avg

def save_checkpoint(model, optimizer, trainloss, testloss, accuracy, epoch, model_dir, args, adv=False):
    output_dir = model_dir
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not adv:
        filename = output_dir +'/normal_train.ckpt'
    else:
        filename = output_dir +'/adv_train_'+args.distance+'.ckpt'

    torch.save({'epoch': epoch,
            'model_state_dict':model.state_dict(),
            'optimizer_state_dict':optimizer.state_dict(),
            'train_loss': trainloss,
            'test_loss': testloss,
            'accuracy': accuracy}, filename)

def test_model(args, device, adv_train=False):
    if args.model_type == "WideResnet":
        model = WideResNet(depth=args.depth, widen_factor=args.widen_factor, num_classes=args.num_classes)
    print(model)
    
    train_loader, test_loader = get_dataloaders(args.batch_size, args.test_batch_size)
    model.to(device)
    model_path = args.model_dir +'/adv_train_l_inf.ckpt'
    model.load_state_dict(torch.load(model_path)['model_state_dict'])

    eval_adv_test_whitebox(model, device, test_loader, args)

def train_model(args, device, adv_train=False):
    if args.model_type == "WideResnet":
        model = WideResNet(depth=args.depth, widen_factor=args.widen_factor, num_classes=args.num_classes)
    else:
        model = ResNet(BasicBlock, args.structure, num_classes=args.num_classes)
    print(model)
    

    train_loader, test_loader = get_dataloaders(args.batch_size, args.test_batch_size)
    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum, weight_decay=args.weight_decay)
    criterion = nn.CrossEntropyLoss()

    if adv_train:
        for epoch in range(args.total_epochs):
            adjust_learning_rate(optimizer, epoch, args)
            loss = adversarial_train(train_loader, model, device, criterion, optimizer, epoch, args)
            testloss, accuracy = test(model,device,criterion,test_loader)
        
        save_checkpoint(model, optimizer, loss, testloss, accuracy, epoch, args.model_dir, args, adv=True)
        eval_adv_test_whitebox(model, device, test_loader, args)
    else:
        for epoch in range(args.total_epochs):
            adjust_learning_rate(optimizer, epoch, args)
            loss = train(train_loader, model, device, criterion, optimizer, epoch, args)
            testloss, accuracy = test(model,device,criterion,test_loader)
        save_checkpoint(model, optimizer, loss, testloss, accuracy, epoch, args.model_dir, args, adv=False)
        eval_adv_test_whitebox(model, device, test_loader, args)
       
def main():
    parser = ArgumentParser(description='PyTorch CIFAR10 Example')
    parser.add_argument('--batch-size', type=int, default=128, metavar='N',
                        help='input batch size for training (default: 128 for cifar10)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--num-classes', type=int, default=10, metavar='N',
                        help='number of class (default: 10)')
    parser.add_argument('--widen_factor', type=int, default=10, metavar='N',
                        help='widen_factor (default: 10)')
    parser.add_argument('--depth', type=int, default=34, metavar='N',
                        help='depth (default: 34)')
    parser.add_argument('--model_type', type=str, default="WideResnet", metavar='N',
                        help='model type (default: WideResnet)')
    parser.add_argument('--structure', nargs='+', type=int, default=[1, 1])
    parser.add_argument('--model_dir', type = str)
    parser.add_argument('--random',
                    default=True,
                    help='random initialization for PGD')

    parser.add_argument('--total_epochs', type=int, default=76, metavar='N',
                        help='number of epochs to train (default: 76)')
    parser.add_argument('--lr', type=float, default=0.1, metavar='LR',
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.9, metavar='M',
                        help='SGD momentum (default: 0.5)')
    parser.add_argument('--weight-decay', '--wd', default=2e-4,
                    type=float, metavar='W')
    parser.add_argument('--TRADES', action='store_true', default=False,
                        help='enable adversarial training using TRADES')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--adv_train', action='store_true', default=False,
                        help='enable adversarial training')
    '''
    CIFAR10 setup. We apply ResNet-18 [HZRS16] for classification. The output size of the last layer is 10. We
    set perturbation  = 0.031, perturbation step size η1 = 0.007, number of iterations K = 10, learning rate
    η2 = 0.1, batch size m = 128, and run 100 epochs on the training dataset. To evaluate the robust error, we
    apply FGSMk
    (white-box) attack with 20 iterations and the step size is 0.003
    https://arxiv.org/pdf/1901.08573.pdf
    '''
    parser.add_argument('--epsilon', type=float, default=0.031,
                        help='perturbation')
    parser.add_argument('--num_steps', type=float, default=10,
                        help='perturb number of steps')
    parser.add_argument('--step_size', type=float, default=0.007,
                        help='perturb step size')

    parser.add_argument('--attack_num_steps', type=int, default=20,
                        help='perturb number of steps')
    parser.add_argument('--attack_step_size', type=float, default=0.003,
                        help='perturb step size')

    parser.add_argument('--beta', type=float, default=6.0,
                        help='regularization, i.e., 1/lambda in TRADES')
    parser.add_argument('--distance', default='l_inf',
                        help='l_2, l_inf')

    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=100, metavar='N',
                        help='how many batches to wait before logging training status')

    args = parser.parse_args()

    print('CIFAR10')
    print(args)
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if use_cuda else "cpu")

    fix_seeds(args.seed)

    if args.distance == 'l_inf':
        print("===========Test Adversarial Linf ===========")
        test_model(args, device, adv_train=True)
        print("===========End L_inf Test ===========\n\n\n")

    print(args.model_dir)
    

if __name__ == '__main__':
    main()



